@@ -1,15 +1,13 @@
 #include "bdb.h"
 
 static void
-mark_lsn(lsnst)
-    struct dblsnst *lsnst;
+mark_lsn(struct dblsnst *lsnst)
 {
     rb_gc_mark(lsnst->env);
 }
 
 static void
-free_lsn(lsnst)
-    struct dblsnst *lsnst;
+free_lsn(struct dblsnst *lsnst)
 {
     if (BDB_VALID(lsnst->env, T_DATA)) {
 	bdb_clean_env(lsnst->env, lsnst->self);
@@ -30,8 +28,7 @@ free_lsn(lsnst)
 }
 
 VALUE
-bdb_makelsn(env)
-    VALUE env;
+bdb_makelsn(VALUE env)
 {
     bdb_ENV *envst;
     struct dblsnst *lsnst;
@@ -42,14 +39,11 @@ bdb_makelsn(env)
     lsnst->env = env;
     lsnst->lsn = ALLOC(DB_LSN);
     lsnst->self = res;
-    rb_ary_push(envst->db_ary, res);
     return res;
 }
 
 static VALUE
-bdb_s_log_put_internal(obj, a, flag)
-    VALUE obj, a;
-    int flag;
+bdb_s_log_put_internal(VALUE obj, VALUE a, int flag)
 {
     bdb_ENV *envst;
     VALUE ret;
@@ -61,7 +55,7 @@ bdb_s_log_put_internal(obj, a, flag)
     ret = bdb_makelsn(obj);
     Data_Get_Struct(ret, struct dblsnst, lsnst);
     data.data = StringValuePtr(a);
-    data.size = RSTRING(a)->len;
+    data.size = RSTRING_LEN(a);
 #if BDB_VERSION < 30000
     if (!envst->envp->lg_info) {
 	rb_raise(bdb_eFatal, "log region not open");
@@ -78,10 +72,7 @@ bdb_s_log_put_internal(obj, a, flag)
 }
 
 static VALUE
-bdb_s_log_put(argc, argv, obj)
-    int argc;
-    VALUE *argv;
-    VALUE obj;
+bdb_s_log_put(int argc, VALUE *argv, VALUE obj)
 {
     VALUE a, b;
     int flag = 0;
@@ -99,8 +90,7 @@ bdb_s_log_put(argc, argv, obj)
 }
 
 static VALUE
-bdb_s_log_checkpoint(obj, a)
-    VALUE obj, a;
+bdb_s_log_checkpoint(VALUE obj, VALUE a)
 {
 #ifdef DB_CHECKPOINT
     return bdb_s_log_put_internal(obj, a, DB_CHECKPOINT);
@@ -111,9 +101,7 @@ bdb_s_log_checkpoint(obj, a)
 }
 
 static VALUE
-bdb_s_log_flush(argc, argv, obj)
-    int argc;
-    VALUE obj, *argv;
+bdb_s_log_flush(int argc, VALUE *argv, VALUE obj)
 {
     bdb_ENV *envst;
 
@@ -142,8 +130,7 @@ bdb_s_log_flush(argc, argv, obj)
 }
 
 static VALUE
-bdb_s_log_curlsn(obj, a)
-    VALUE obj, a;
+bdb_s_log_curlsn(VALUE obj, VALUE a)
 {
 #ifdef DB_CURSLN
     return bdb_s_log_put_internal(obj, a, DB_CURLSN);
@@ -155,9 +142,7 @@ bdb_s_log_curlsn(obj, a)
   
 
 static VALUE
-bdb_env_log_stat(argc, argv, obj)
-    int argc;
-    VALUE obj, *argv;
+bdb_env_log_stat(int argc, VALUE *argv, VALUE obj)
 {
     DB_LOG_STAT *bdb_stat;
     bdb_ENV *envst;
@@ -236,8 +221,7 @@ bdb_env_log_stat(argc, argv, obj)
 #if BDB_VERSION < 40000
 
 static VALUE
-bdb_env_log_get(obj, a)
-    VALUE obj, a;
+bdb_env_log_get(VALUE obj, VALUE a)
 {
     bdb_ENV *envst;
     DBT data;
@@ -273,29 +257,40 @@ bdb_env_log_get(obj, a)
 static VALUE bdb_log_cursor _((VALUE));
 #endif
 
+#define BDB_LOG_INIT 0
+#define BDB_LOG_SET 1
+#define BDB_LOG_NEXT 2
+
 static VALUE
-bdb_i_each_log_get(obj, flag)
-    VALUE obj;
-    int flag;
+bdb_i_each_log_get(VALUE obj, int flag)
 {
+#if BDB_VERSION < 40000
     bdb_ENV *envst;
-    struct dblsnst *lsnst;
+#endif
+    struct dblsnst *lsnst, *lsnst1;
     DBT data;
-    VALUE lsn, res;
+    VALUE lsn, res, lsn1;
     int ret, init, flags;
 
+    init = BDB_LOG_INIT;
 #if BDB_VERSION < 40000
     GetEnvDB(obj, envst);
 #else
     lsn = obj;
     Data_Get_Struct(obj, struct dblsnst, lsnst);
-    if (lsnst->cursor == 0) {
-	bdb_log_cursor(obj);
-    }
     flag = lsnst->flags;
+    if (lsnst->cursor == 0) {
+	DB_LSN *lsn1;
+
+	init = BDB_LOG_SET;
+	lsn1 = lsnst->lsn;
+	lsn = bdb_makelsn(lsnst->env);
+	Data_Get_Struct(lsn, struct dblsnst, lsnst);
+	MEMCPY(lsnst->lsn, lsn1, DB_LSN, 1);
+	bdb_log_cursor(lsn);
+    }
 #endif
 
-    init = 0; /* strange ??? */
     do {
 #if BDB_VERSION < 40000
 	lsn = bdb_makelsn(obj);
@@ -303,12 +298,18 @@ bdb_i_each_log_get(obj, flag)
 #endif
 	MEMZERO(&data, DBT, 1);
 	data.flags |= DB_DBT_MALLOC;
-	if (!init) {
+	switch (init) {
+	case BDB_LOG_INIT:
 	    flags = (flag == DB_NEXT)?DB_FIRST:DB_LAST;
-	    init = 1;
-	}
-	else
+	    break;
+	case BDB_LOG_SET:
+	    flags = DB_SET;
+	    break;
+	default:
 	    flags = flag;
+	    break;
+	}
+	init = BDB_LOG_NEXT;
 #if BDB_VERSION < 30000
 	if (!envst->envp->lg_info) {
 	    rb_raise(bdb_eFatal, "log region not open");
@@ -317,8 +318,12 @@ bdb_i_each_log_get(obj, flag)
 #else
 #if BDB_VERSION >= 40000
 	ret = bdb_test_error(lsnst->cursor->get(lsnst->cursor, lsnst->lsn, &data, flags));
+	lsn1 = bdb_makelsn(lsnst->env);
+	Data_Get_Struct(lsn1, struct dblsnst, lsnst1);
+	MEMCPY(lsnst1->lsn, lsnst->lsn, DB_LSN, 1);
 #else
 	ret = bdb_test_error(log_get(envst->envp, lsnst->lsn, &data, flags));
+	lsn1 = lsn;
 #endif
 #endif
 	if (ret == DB_NOTFOUND) {
@@ -334,15 +339,13 @@ bdb_i_each_log_get(obj, flag)
 #if BDB_VERSION < 40000
  
 static VALUE
-bdb_env_log_each(obj) 
-    VALUE obj;
+bdb_env_log_each(VALUE obj) 
 { 
     return bdb_i_each_log_get(obj, DB_NEXT);
 }
 
 static VALUE
-bdb_env_log_hcae(obj)
-    VALUE obj;
+bdb_env_log_hcae(VALUE obj)
 { 
     return bdb_i_each_log_get(obj, DB_PREV);
 }
@@ -350,8 +353,7 @@ bdb_env_log_hcae(obj)
 #else
 
 static VALUE
-log_cursor_close(obj)
-    VALUE obj;
+log_cursor_close(VALUE obj)
 {
     struct dblsnst *lsnst;
 
@@ -364,39 +366,39 @@ log_cursor_close(obj)
 }
 
 static VALUE
-bdb_log_cursor_close(obj)
-    VALUE obj;
+bdb_log_cursor_close(VALUE obj)
 {
     struct dblsnst *lsnst;
 
     Data_Get_Struct(obj, struct dblsnst, lsnst);
     bdb_clean_env(lsnst->env, obj);
+    
     return log_cursor_close(obj);
 }
 
 static VALUE
-bdb_log_cursor(lsn)
-    VALUE lsn;
+bdb_log_cursor(VALUE lsn)
 {
     bdb_ENV *envst;
     struct dblsnst *lsnst;
 
     Data_Get_Struct(lsn, struct dblsnst, lsnst);
-    GetEnvDB(lsnst->env, envst);
-    bdb_test_error(envst->envp->log_cursor(envst->envp, &lsnst->cursor, 0));
+    if (!lsnst->cursor) {
+	GetEnvDB(lsnst->env, envst);
+	bdb_test_error(envst->envp->log_cursor(envst->envp, &lsnst->cursor, 0));
+	bdb_ary_push(&envst->db_ary, lsn);
+    }
     return lsn;
 }
     
 static VALUE
-bdb_env_log_cursor(obj)
-    VALUE obj;
+bdb_env_log_cursor(VALUE obj)
 {
     return bdb_log_cursor(bdb_makelsn(obj));
 }
 
 static VALUE
-bdb_log_i_get(obj)
-    VALUE obj;
+bdb_env_i_get(VALUE obj)
 {
     bdb_ENV *envst;
     struct dblsnst *lsnst;
@@ -408,20 +410,9 @@ bdb_log_i_get(obj)
     return bdb_i_each_log_get(obj, lsnst->flags);
 }
 
-static VALUE
-bdb_log_each(lsn)
-    VALUE lsn;
-{
-    struct dblsnst *lsnst;
-
-    Data_Get_Struct(lsn, struct dblsnst, lsnst);
-    lsnst->flags = DB_NEXT;
-    return bdb_log_i_get(lsn);
-}
 
 static VALUE
-bdb_env_log_each(obj)
-    VALUE obj;
+bdb_env_log_each(VALUE obj)
 {
     VALUE lsn;
     struct dblsnst *lsnst;
@@ -429,23 +420,11 @@ bdb_env_log_each(obj)
     lsn = bdb_makelsn(obj);
     Data_Get_Struct(lsn, struct dblsnst, lsnst);
     lsnst->flags = DB_NEXT;
-    return rb_ensure(bdb_log_i_get, lsn, bdb_log_cursor_close, lsn);
+    return rb_ensure(bdb_env_i_get, lsn, bdb_log_cursor_close, lsn);
 }
 
 static VALUE
-bdb_log_hcae(lsn)
-    VALUE lsn;
-{
-    struct dblsnst *lsnst;
-
-    Data_Get_Struct(lsn, struct dblsnst, lsnst);
-    lsnst->flags = DB_PREV;
-    return bdb_log_i_get(lsn);
-}
-
-static VALUE
-bdb_env_log_hcae(obj)
-    VALUE obj;
+bdb_env_log_hcae(VALUE obj)
 {
     VALUE lsn;
     struct dblsnst *lsnst;
@@ -453,15 +432,44 @@ bdb_env_log_hcae(obj)
     lsn = bdb_makelsn(obj);
     Data_Get_Struct(lsn, struct dblsnst, lsnst);
     lsnst->flags = DB_PREV;
+    return rb_ensure(bdb_env_i_get, lsn, bdb_log_cursor_close, lsn);
+}
+
+static VALUE
+bdb_log_i_get(VALUE obj)
+{
+    struct dblsnst *lsnst;
+
+    log_cursor_close(obj);
+    Data_Get_Struct(obj, struct dblsnst, lsnst);
+    return bdb_i_each_log_get(obj, lsnst->flags);
+}
+
+static VALUE
+bdb_log_each(VALUE lsn)
+{
+    struct dblsnst *lsnst;
+
+    Data_Get_Struct(lsn, struct dblsnst, lsnst);
+    lsnst->flags = DB_NEXT;
     return rb_ensure(bdb_log_i_get, lsn, bdb_log_cursor_close, lsn);
 }
 
+static VALUE
+bdb_log_hcae(VALUE lsn)
+{
+    struct dblsnst *lsnst;
+
+    Data_Get_Struct(lsn, struct dblsnst, lsnst);
+    lsnst->flags = DB_PREV;
+    return rb_ensure(bdb_log_i_get, lsn, bdb_log_cursor_close, lsn);
+ }
+ 
+
 #endif
  
 static VALUE
-bdb_env_log_archive(argc, argv, obj)
-    int argc;
-    VALUE *argv, obj;
+bdb_env_log_archive(int argc, VALUE *argv, VALUE obj)
 {
     char **list, **file;
     bdb_ENV *envst;
@@ -505,8 +513,7 @@ bdb_env_log_archive(argc, argv, obj)
 }
 
 static VALUE
-bdb_lsn_env(obj)
-    VALUE obj;
+bdb_lsn_env(VALUE obj)
 {
     struct dblsnst *lsnst;
     bdb_ENV *envst;
@@ -515,8 +522,7 @@ bdb_lsn_env(obj)
 }
 
 static VALUE
-bdb_lsn_log_file(obj)
-    VALUE obj;
+bdb_lsn_log_file(VALUE obj)
 {
     struct dblsnst *lsnst;
     bdb_ENV *envst;
@@ -539,8 +545,7 @@ bdb_lsn_log_file(obj)
 }
 
 static VALUE
-bdb_lsn_log_flush(obj)
-    VALUE obj;
+bdb_lsn_log_flush(VALUE obj)
 {
     struct dblsnst *lsnst;
     bdb_ENV *envst;
@@ -562,8 +567,7 @@ bdb_lsn_log_flush(obj)
 }
 
 static VALUE
-bdb_lsn_log_compare(obj, a)
-    VALUE obj, a;
+bdb_lsn_log_compare(VALUE obj, VALUE a)
 {
     struct dblsnst *lsnst1, *lsnst2;
     bdb_ENV *envst1, *envst2;
@@ -577,27 +581,24 @@ bdb_lsn_log_compare(obj, a)
 }
 
 static VALUE
-bdb_lsn_log_get(argc, argv, obj)
-    int argc;
-    VALUE obj, *argv;
+bdb_lsn_log_get(int argc, VALUE *argv, VALUE obj)
 {
     struct dblsnst *lsnst;
-    bdb_ENV *envst;
     DBT data;
     VALUE res, a;
     int ret, flags;
+    bdb_ENV *envst;
+#if BDB_VERSION >= 40000
+    DB_LOGC *cursor;
+#endif
 
     flags = DB_SET;
     if (rb_scan_args(argc, argv, "01", &a) == 1) {
 	flags = NUM2INT(a);
     }
-#if BDB_VERSION < 40000
     GetLsn(obj, lsnst, envst);
-#else
-    Data_Get_Struct(obj, struct dblsnst, lsnst);
-    if (lsnst->cursor == 0) {
-	bdb_log_cursor(obj);
-    }
+#if BDB_VERSION >= 40000
+    bdb_test_error(envst->envp->log_cursor(envst->envp, &cursor, 0));
 #endif
     MEMZERO(&data, DBT, 1);
     data.flags |= DB_DBT_MALLOC;
@@ -608,7 +609,9 @@ bdb_lsn_log_get(argc, argv, obj)
     ret = bdb_test_error(log_get(envst->envp->lg_info, lsnst->lsn, &data, flags));
 #else
 #if BDB_VERSION >= 40000
-    ret = bdb_test_error(lsnst->cursor->get(lsnst->cursor, lsnst->lsn, &data, flags));
+    ret = cursor->get(cursor, lsnst->lsn, &data, flags);
+    cursor->close(cursor, 0);
+    ret = bdb_test_error(ret);
 #else
     ret = bdb_test_error(log_get(envst->envp, lsnst->lsn, &data, flags));
 #endif
@@ -622,8 +625,7 @@ bdb_lsn_log_get(argc, argv, obj)
 }
 
 static VALUE
-bdb_log_register(obj, a)
-    VALUE obj, a;
+bdb_log_register(VALUE obj, VALUE a)
 {
 #if BDB_VERSION >= 40100
     rb_warn("log_register is obsolete");
@@ -661,8 +663,7 @@ bdb_log_register(obj, a)
 }
 
 static VALUE
-bdb_log_unregister(obj)
-    VALUE obj;
+bdb_log_unregister(VALUE obj)
 {
 #if BDB_VERSION >= 40100
     rb_warn("log_unregister is obsolete");
@@ -715,7 +716,11 @@ void bdb_init_log()
     rb_define_method(bdb_cCommon, "log_unregister", bdb_log_unregister, 0);
     bdb_cLsn = rb_define_class_under(bdb_mDb, "Lsn", rb_cObject);
     rb_include_module(bdb_cLsn, rb_mComparable);
+#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
+    rb_undef_alloc_func(bdb_cLsn);
+#else
     rb_undef_method(CLASS_OF(bdb_cLsn), "allocate");
+#endif
     rb_undef_method(CLASS_OF(bdb_cLsn), "new");
     rb_define_method(bdb_cLsn, "env", bdb_lsn_env, 0);
 #if BDB_VERSION >= 40000
@@ -728,7 +733,7 @@ void bdb_init_log()
     rb_define_method(bdb_cLsn, "log_reverse_each", bdb_log_hcae, 0);
     rb_define_method(bdb_cLsn, "reverse_each", bdb_log_hcae, 0);
 #endif
-    rb_define_method(bdb_cLsn, "log_get", bdb_lsn_log_get, -1);
+     rb_define_method(bdb_cLsn, "log_get", bdb_lsn_log_get, -1);
     rb_define_method(bdb_cLsn, "get", bdb_lsn_log_get, -1);
     rb_define_method(bdb_cLsn, "log_compare", bdb_lsn_log_compare, 1);
     rb_define_method(bdb_cLsn, "compare", bdb_lsn_log_compare, 1);
